/* big.c: high-speed LISP interpreter */

/*
   The storage required by this interpreter is 8 bytes times
   the symbolic constant SIZE, which is 8 * 64,000,000 =
   512 megabytes.  To run this interpreter in small machines,
   reduce the #define SIZE 64000000 below.

   To compile, type
      cc -O -obig -bmaxdata:0x40000000 big.c
   To run interactively, type
      big
   To run with output on screen, type
      big <test.l
   To run with output in file, type
      big <test.l >test.r

   Reference:  Kernighan & Ritchie,
   The C Programming Language, Second Edition,
   Prentice-Hall, 1988.
*/

#include <stdio.h>
#include <time.h>

#define SIZE 64000000 /* numbers of nodes of tree storage */
#define LAST_ATOM 128 /* highest integer value of character */
#define nil 128 /* null pointer in tree storage */
#define question -1 /* error pointer in tree storage */
#define exclamation -2 /* error pointer in tree storage */
#define infinity 999999999 /* "infinite" depth limit */

/* For very small PC's, change following line to
   make hd & tl unsigned short instead of long:
   (If so, SIZE must be less than 64K.) */
long hd[SIZE+1], tl[SIZE+1]; /* tree storage */
long next = nil; /* list of free nodes */
long low = LAST_ATOM+1; /* first never-used node */
long vlst[LAST_ATOM+1]; /* bindings of each atom */
long tape; /* Turing machine tapes */
long display; /* display indicators */
long outputs; /* output stacks */
long q; /* for converting expressions to binary */
long col; /* column in each 50 character chunk of output
            (preceeded by 12 char prefix) */
long cc; /* character count */
time_t time1; /* clock at start of execution */
time_t time2; /* clock at end of execution */

long ev(long e); /* initialize and evaluate expression */
void initialize_atoms(void); /* initialize atoms */
void clean_env(void); /* clean environment */
void restore_env(void); /* restore dirty environment */
long eval(long e, long d); /* evaluate expression */
/* evaluate list of expressions */
long evalst(long e, long d);
/* bind values of arguments to formal parameters */
void bind(long vars, long args);
long at(long x); /* atomic predicate */
long jn(long x, long y); /* join head to tail */
long pop(long x); /* return tl & free node */
void fr(long x); /* free list of nodes */
long eq(long x, long y); /* equal predicate */
long cardinality(long x); /* number of elements in list */
long append(long x, long y); /* append two lists */
/* read one square of Turing machine tape */
long getbit(void);
/* read one character from Turing machine tape */
long getchr(void);
/* read expression from Turing machine tape */
void putchr(long x); /* convert character to binary */
void putexp(long x); /* convert expression to binary */
void putexp2(long x); /* convert expression to binary */
long out(char *x, long y); /* output expression */
void out2(long x); /* really output expression */
void out3(long x); /* really really output expression */
long chr2(void); /* read character - skip blanks,
                    tabs and new line characters */
long chr(void); /* read character - skip comments */
long in(long mexp, long rparenokay); /* input m-exp */
long (*pchr)(void); /* pointer to chr2 or getchr */

main() /* lisp main program */
{
char name_colon[] = "X:"; /* for printing name: def pairs */

time1 = time(NULL); /* start timer */
printf("big.c\n\nLISP Interpreter Run\n");
initialize_atoms();

while (1) {
      long e, f, name, def;
      printf("\n");
      /* read lisp meta-expression from stdin, ) not okay */
      pchr = chr2;
      e = in(1,0);
      /* flush rest of input line */
      while (putchar(getchar()) != '\n');
      printf("\n");
      f = hd[e];
      name = hd[tl[e]];
      def = hd[tl[tl[e]]];
      if (f == '&') {
      /* definition */
         if (at(name)) {
         /* variable definition, e.g., & x '(abc) */
            def = out("expression",def);
            def = ev(def);
         } /* end of variable definition */
         else          {
         /* function definition, e.g., & (Fxy) *x*y() */
            long var_list = tl[name];
            name = hd[name];
            def = jn('&',jn(var_list,jn(def,nil)));
         } /* end of function definition */
         name_colon[0] = name;
         out(name_colon,def);
         /* new binding replaces old */
         vlst[name] = jn(def,nil);
         continue;
      } /* end of definition */
      /* write corresponding s-expression */
      e = out("expression",e);
      /* evaluate expression */
      e = out("value",ev(e));
   }
}

long ev(long e) /* initialize and evaluate expression */
{
 long d = infinity; /* "infinite" depth limit */
 long v;
 tape = jn(nil,nil);
 display = jn('Y',nil);
 outputs = jn(nil,nil);
 v = eval(e,d);
 if (v == question) v = '?';
 if (v == exclamation) v = '!';
 return v;
}

void initialize_atoms(void) /* initialize atoms */
{
 long i;
 for (i = 0; i <= LAST_ATOM; ++i) {
 hd[i] = tl[i] = i; /* so that hd & tl of atom = atom */
 /* initially each atom evaluates to self */
 vlst[i] = jn(i,nil);
 }
}

long jn(long x, long y) /* join two lists */
{
 long z;
 /* if y is not a list, then jn is x */
 if ( y != nil && at(y) ) return x;

 if (next == nil) {
  if (low > SIZE) {
  printf("Storage overflow!\n");
  exit(0);
  }
 next = low++;
 tl[next] = nil;
 }

 z = next;
 next = tl[next];
 hd[z] = x;
 tl[z] = y;

 return z;
}

long pop(long x) /* return tl & free node */
{
 long y;
 y = tl[x];
 tl[x] = next;
 next = x;
 return y;
}

void fr(long x) /* free list of nodes */
{
 while (x != nil) x = pop(x);
}

long at(long x) /* atom predicate */
{
 return ( x <= LAST_ATOM );
}

long eq(long x, long y) /* equal predicate */
{
 if (x == y) return 1;
 if (at(x)) return 0;
 if (at(y)) return 0;
 if (eq(hd[x],hd[y])) return eq(tl[x],tl[y]);
 return 0;
}

long eval(long e, long d) /* evaluate expression */
{
/*
 e is expression to be evaluated
 d is permitted depth - integer, not pointer to tree storage
*/
 long f, v, args, x, y, z, vars, body;

 /* find current binding of atomic expression */
 if (at(e)) return hd[vlst[e]];

 f = eval(hd[e],d); /* evaluate function */
 e = tl[e]; /* remove function from list of arguments */
 if (f < 0) return f; /* function = error value? */

 if (f == '\'') return hd[e]; /* quote */

 if (f == '/') { /* if then else */
 v = eval(hd[e],d);
 e = tl[e];
 if (v < 0) return v; /* error? */
 if (v == '0') e = tl[e];
 return eval(hd[e],d);
 }

 args = evalst(e,d); /* evaluate list of arguments */
 if (args < 0) return args; /* error? */

 x = hd[args]; /* pick up first argument */
 y = hd[tl[args]]; /* pick up second argument */
 z = hd[tl[tl[args]]]; /* pick up third argument */

 switch (f) {
 case '@': {fr(args); return getbit();}
         /* read lisp meta-expression from TM tape, ) not okay */
 case '%': {fr(args); pchr = getchr; return in(1,0);}
 case '#': {fr(args);
           v = q = jn(nil,nil); putexp(x); return pop(v);}
 case '+': {fr(args); return hd[x];}
 case '-': {fr(args); return tl[x];}
 case '.': {fr(args); return (at(x) ? '1' : '0');}
 case ',': {fr(args); hd[outputs] = jn(x,hd[outputs]);
           return (hd[display] == 'Y' ? out("display",x): x);}
 case '~': {fr(args); return /* out("show", */ x /* ) */ ;}
 case '=': {fr(args); return (eq(x,y) ? '1' : '0');}
 case '*': {fr(args); return jn(x,y);}
 case '^': {fr(args);
           return append((at(x)?nil:x),(at(y)?nil:y));}
 }

 if (d == 0) {fr(args); return question;} /* depth exceeded
                                             -> error! */
 d--; /* decrement depth */

 if (f == '!') {
 fr(args);
 clean_env(); /* clean environment */
 v = eval(x,d);
 restore_env(); /* restore unclean environment */
 return v;
 }

 if (f == '?') {
 fr(args);
 x = cardinality(x); /* convert s-exp into number */
 clean_env();
 tape = jn(z,tape);
 display = jn('N',display);
 outputs = jn(nil,outputs);
 v = eval(y,(d <= x ? d : x));
 restore_env();
 z = hd[outputs];
 tape = pop(tape);
 display = pop(display);
 outputs = pop(outputs);
 if (v == question) return (d <= x ? question : jn('?',z));
 if (v == exclamation) return jn('!',z);
 return jn(jn(v,nil),z);
 }

 f = tl[f];
 vars = hd[f];
 f = tl[f];
 body = hd[f];

 bind(vars,args);
 fr(args);

 v = eval(body,d);

 /* unbind */
 while (!at(vars)) {
 if (at(hd[vars]))
 vlst[hd[vars]] = pop(vlst[hd[vars]]);
 vars = tl[vars];
 }

 return v;
}

void clean_env(void) /* clean environment */
{
 long i;
 for (i = 0; i <= LAST_ATOM; ++i)
 vlst[i] = jn(i,vlst[i]); /* clean environment */
}

void restore_env(void) /* restore unclean environment */
{
 long i;
 for (i = 0; i <= LAST_ATOM; ++i)
 vlst[i] = pop(vlst[i]); /* restore unclean environment */
}

long cardinality(long x) /* number of elements in list */
{
 if (at(x)) return (x == nil ? 0 : infinity);
 return 1+cardinality(tl[x]);
}

/* bind values of arguments to formal parameters */
void bind(long vars, long args)
{
 if (at(vars)) return;
 bind(tl[vars],tl[args]);
 if (at(hd[vars]))
 vlst[hd[vars]] = jn(hd[args],vlst[hd[vars]]);
}

long evalst(long e, long d) /* evaluate list of expressions */
{
 long x, y;
 if (at(e)) return nil;
 x = eval(hd[e],d);
 if (x < 0) return x; /* error? */
 y = evalst(tl[e],d);
 if (y < 0) return y; /* error? */
 return jn(x,y);
}

long append(long x, long y) /* append two lists */
{
 if (at(x)) return y;
 return jn(hd[x],append(tl[x],y));
}

/* read one square of Turing machine tape */
long getbit(void)
{
 long x;
 if (at(hd[tape])) return exclamation; /* tape finished ! */
 x = hd[hd[tape]];
 hd[tape] = tl[hd[tape]];
 return (x == '0' ? '0' : '1');
}

/* read one character from Turing machine tape */
long getchr(void)
{
 long c, b, i;
 c = 0;
 do {
    for (i = 0; i < 7; ++i) {
    b = getbit();
    if (b < 0) return b; /* error? */
    c = c + c + b - '0';
    }
 }
/* keep only non-blank printable ASCII codes */
 while (c >= 127 || c <= 32) ;
 return c;
}

void putchr(long x) /* convert character to binary */
{
 q = tl[q] = jn(( x &  64 ? '1' : '0' ), nil);
 q = tl[q] = jn(( x &  32 ? '1' : '0' ), nil);
 q = tl[q] = jn(( x &  16 ? '1' : '0' ), nil);
 q = tl[q] = jn(( x &   8 ? '1' : '0' ), nil);
 q = tl[q] = jn(( x &   4 ? '1' : '0' ), nil);
 q = tl[q] = jn(( x &   2 ? '1' : '0' ), nil);
 q = tl[q] = jn(( x &   1 ? '1' : '0' ), nil);
}

void putexp(long x) /* convert expression to binary */
{
 /* remove containing parens at top level! */
 while (!at(x)) {
 putexp2(hd[x]);
 x = tl[x];
 }
}

void putexp2(long x) /* convert expression to binary */
{
 if ( at(x) && x != nil ) {putchr(x); return;}
 putchr('(');

 while (!at(x)) {
 putexp2(hd[x]);
 x = tl[x];
 }

 putchr(')');
}

long out(char *x, long y) /* output expression */
{
   printf("%-12s",x);
   col = 0; /* so can insert \n and 12 blanks
               every 50 characters of output */
   cc = -2; /* count characters in m-expression */
   out2(y);
   printf("\n");
   if (*x == 's' && cc > 0)
   printf("%-12s%d(%o)/%d(%o)\n","size",cc,cc,7*cc,7*cc);
   return y;
}

void out2(long x) /* really output expression */
{
   if ( at(x) && x != nil ) {out3(x); return;}
   out3('(');
   while (!at(x)) {
      out2(hd[x]);
      x = tl[x];
      }
   out3(')');
}

void out3(long x) /* really really output expression */
{
   if (col++ == 50) {printf("\n%-12s"," ");  col = 1;}
   putchar(x);
   cc++;
}

long chr2(void) /* read character - skip blanks,
                   tabs and new line characters */
{
   long c;
   do {
      c = getchar();
      if (c == EOF) {
         time2 = time(NULL);
         printf(
         "End of LISP Run\n\nElapsed time is %.0f seconds.\n",
         difftime(time2,time1)
      /* on some systems, above line should instead be: */
      /* time2 - time1 */
         );
         exit(0); /* terminate execution */
         }
      putchar(c);
   }
/* keep only non-blank printable ASCII codes */
   while (c >= 127 || c <= 32) ;
   return c;
}

long chr(void) /* read character - skip comments */
{              /* here pchr -> chr2, getchr */
   long c;
   while (1) {
      c = (*pchr)();
      if (c < 0) return c; /* error? */
      if (c != '[') return c;
   /* comments may be nested */
      while ((c = chr()) != ']')
      if (c < 0) return c; /* error? */
   }
}

/* read expression from Turing machine tape */
long in(long mexp, long rparenokay) /* input m-exp */
{
   long c = chr();
   if (c < 0) return c; /* error? */
   if (c == ')') if (rparenokay) return ')'; else return nil;
   if (c == '(') { /* explicit list */
      long first, last, next;
      first = last = jn(nil,nil);
      while ((next = in(mexp,1)) != ')')
      {
      if (next < 0) return next; /* error? */
      last = tl[last] = jn(next,nil);
      }
      return pop(first);
      }
   if (!mexp) return c; /* atom */
   if (c == '{') { /* number */
      long n = 0, u;
      while ((c = chr()) != '}') {
      if (c < 0) return c; /* error? */
      c = c - '0';
      if (c >= 0 && c <= 9) n = 10 * n + c;
      }
      for (u = nil; n > 0; n--) u = jn('1',u);
      return u;
      }
   if (c == '"') return in(0,0); /* s-exp */
   if (c == ':') { /* expand "let" */
      long name, def, body;
      name = in(1,0); if (name < 0) return name; /* error? */
      def  = in(1,0); if (def  < 0) return def ; /* error? */
      body = in(1,0); if (body < 0) return body; /* error? */
      if (!at(name)) {
         long var_list;
         var_list = tl[name];
         name = hd[name];
         def =
         jn('\'',jn(jn('&',jn(var_list,jn(def,nil))),nil));
         }
      return
      jn(
       jn('\'',jn(jn('&',jn(jn(name,nil),jn(body,nil))),nil)),
       jn(def,nil)
      );
      }
   switch (c) { long x, y, z;
      case '@': case '%':
         return jn(c,nil);
      case '+': case '-': case '.': case '\'':
      case ',': case '!': case '#': case '~':
        {x = in(1,0); if (x < 0) return x; /* error? */
         return jn(c,jn(x,nil));}
      case '*': case '=': case '&': case '^':
        {x = in(1,0); if (x < 0) return x; /* error? */
         y = in(1,0); if (y < 0) return y; /* error? */
         return jn(c,jn(x,jn(y,nil)));}
      case '/': case ':': case '?':
        {x = in(1,0); if (x < 0) return x; /* error? */
         y = in(1,0); if (y < 0) return y; /* error? */
         z = in(1,0); if (z < 0) return z; /* error? */
         return jn(c,jn(x,jn(y,jn(z,nil))));}
      default:
         return c;
      }
}
